/*
 * $PostgreSQL: pgsql/contrib/pg_heapdump/pg_heapdump.c,v 1.25 2006/10/04 00:29:46 momjian Exp $
 *
 * Copyright (c) 2007	Takahiro Itagaki
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above
 * copyright notice and this paragraph and the following two
 * paragraphs appear in all copies.
 *
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT,
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS
 * IS" BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include "postgres.h"

#include "fmgr.h"
#include "funcapi.h"
#include "access/heapam.h"
#include "access/xact.h"
#include "catalog/namespace.h"
#include "catalog/pg_type.h"
#include "utils/builtins.h"


PG_MODULE_MAGIC;

extern Datum pg_heapdump_name(PG_FUNCTION_ARGS);
extern Datum pg_heapdump_oid(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(pg_heapdump_name);
PG_FUNCTION_INFO_V1(pg_heapdump_oid);

static Datum pg_heapdump(Oid relid, text *relname, FunctionCallInfo fcinfo);

Datum
pg_heapdump_name(PG_FUNCTION_ARGS)
{
	PG_RETURN_DATUM(pg_heapdump(InvalidOid, PG_GETARG_TEXT_P(0), fcinfo));
}

Datum
pg_heapdump_oid(PG_FUNCTION_ARGS)
{
	PG_RETURN_DATUM(pg_heapdump(PG_GETARG_OID(0), NULL, fcinfo));
}

#define NUM_HEAPDUMP_ELEM	9

/*
 * pg_heapdump -- return set of record meta data.
 */
static Datum
pg_heapdump(Oid relid, text *relname, FunctionCallInfo fcinfo)
{
	FuncCallContext	   *funcctx;
	HeapScanDesc		scan;
	HeapTuple			tuple;

	if (SRF_IS_FIRSTCALL())
	{
		MemoryContext	oldctx;
		Relation		rel;
		TupleDesc		tupledesc;

		if (relname != NULL)
			rel = heap_openrv(makeRangeVarFromNameList(
				textToQualifiedNameList(relname)), AccessShareLock);
		else
			rel = heap_open(relid, AccessShareLock);

		funcctx = SRF_FIRSTCALL_INIT();

		/* Switch scan when allocating stuff to be used in later calls */
		oldctx = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

		/* get a tuple descriptor for our result type */
		if (get_call_result_type(fcinfo, NULL, &tupledesc) != TYPEFUNC_COMPOSITE)
			elog(ERROR, "return type must be a row type");
		funcctx->tuple_desc = BlessTupleDesc(tupledesc);

		/* Create a function scan for cross-call persistence. */
		funcctx->user_fctx = heap_beginscan(rel, SnapshotAny, 0, NULL);

		/* Return to original scan when allocating transient memory */
		MemoryContextSwitchTo(oldctx);
	}

	funcctx = SRF_PERCALL_SETUP();
	scan = (HeapScanDesc) funcctx->user_fctx;

	if ((tuple = heap_getnext(scan, ForwardScanDirection)) != NULL)
	{
		Datum			values[NUM_HEAPDUMP_ELEM];
		bool			nulls[NUM_HEAPDUMP_ELEM] = { 0 };
		HeapTuple		ret;
		BlockNumber		blknum;
		OffsetNumber	offnum;
		Page			page;
		ItemId			itemid;
		HeapTupleHeader	htup = tuple->t_data;

		blknum = ItemPointerGetBlockNumber(&tuple->t_self);
		offnum = ItemPointerGetOffsetNumber(&tuple->t_self);
		page = BufferGetPage(scan->rs_cbuf);
		itemid = PageGetItemId(page, offnum);

		values[0] = Int64GetDatum(blknum);
		values[1] = Int16GetDatum(offnum);
		values[2] = Int16GetDatum(ItemIdGetOffset(itemid));
		values[3] = Int16GetDatum(ItemIdGetLength(itemid));

		if (!(htup->t_infomask & HEAP_XMIN_INVALID))
		{
			TransactionId	xmin = HeapTupleHeaderGetXmin(htup);

			values[4] = Int64GetDatum(xmin);

			if (!(htup->t_infomask & HEAP_MOVED) &&
				TransactionIdIsCurrentTransactionId(xmin))
				values[5] = Int64GetDatum(HeapTupleHeaderGetCmin(htup));
			else
				nulls[5] = true;
		}
		else
		{
			nulls[4] = true;
			nulls[5] = true;
		}

		if (!(htup->t_infomask & HEAP_XMAX_INVALID))
		{
			TransactionId	xmax = HeapTupleHeaderGetXmax(htup);

			values[6] = Int64GetDatum(xmax);

			if (!(htup->t_infomask & (HEAP_MOVED | HEAP_IS_LOCKED)) &&
				TransactionIdIsCurrentTransactionId(xmax))
				values[7] = Int64GetDatum(HeapTupleHeaderGetCmax(htup));
			else
				nulls[7] = true;
		}
		else
		{
			nulls[6] = true;
			nulls[7] = true;
		}

		values[8] = Int32GetDatum(htup->t_infomask);

		/* Build and return the tuple. */
		ret = heap_form_tuple(funcctx->tuple_desc, values, nulls);

		SRF_RETURN_NEXT(funcctx, HeapTupleGetDatum(ret));
	}
	else
	{
		Relation	rel = scan->rs_rd;
		heap_endscan(scan);
		heap_close(rel, AccessShareLock);
		SRF_RETURN_DONE(funcctx);
	}
}
